/*
    gutenfetch - a small utility to list and fetch books available through
	project gutenberg

    Copyright (C) 2001 Russell Francis 

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <stdarg.h>
#include <string.h>
#include <sys/uio.h>
#include <sys/types.h>
#include "gutenfetch.h"
#include "server.h"
#include "../config.h"

#define TRUE  		1
#define FALSE 		0
#define BLOCK_SIZE	8192

static int verbose = TRUE;
static int ventries = 0;
static char **host = NULL;

int
main(argc, argv)
	int argc;
	char **argv;
{
	int ch, len;
	int list = FALSE;
	char *command = NULL;
	char *fetch_file = NULL;
	char *outfile = NULL;
	char *force_host = NULL;

	// load config file get default servers etc...
	LoadConfigFile();

	while ((ch = getopt(argc, argv, "s:o:f:vql")) != -1){
		switch (ch) {
		case 'v':
			verbose = TRUE;
			break;
		case 'q':
			verbose = FALSE;
			break;
		case 'l':	// list all books
			list = TRUE;
			break;
		case 's':
			CleanUp(1, &force_host);
			if(asprintf(&force_host, "%s", optarg) == -1){
				if(verbose)
					fprintf(stderr,
					"gutenfetch.c : main() : memory allocation error: using default\n");
			}
			break;
		case 'o':
			CleanUp(1, &outfile);
			if(asprintf(&outfile, "%s", optarg) == -1){
				if(verbose)
					fprintf(stderr,
					"gutenfetch.c : main() : memory allocation error: using default\n");
			}
			break;
		case 'f':
			CleanUp(1, &fetch_file);
			if(asprintf(&fetch_file, "%s", optarg) == -1){
				if(verbose)
					fprintf(stderr,
					"gutenfetch.c : main() : memory allocation error: using default\n");
			}
			break;
		default:
			PrintUsage();
			CleanUp(4, &fetch_file, &outfile, &force_host, &command);
			exit(TRUE);
			break;	
		}	
	}

	// set the outfile to wget stdout
	if(outfile == NULL){
		if(asprintf(&outfile, "-") == -1){
			if(verbose)
				fprintf(stderr,
				"gutenfetch.c : main() : critical memory allocation error\n");
			CleanUp(4, &fetch_file, &outfile, &force_host, &command);
			return FALSE;
		}
	}

	if(list){ // fetch and print formated GUTINDEX.ALL off of server of choice
		// build command
		if(force_host != NULL){
			AssembleCommand(&command, force_host, "GUTINDEX.ALL", "-");
			FetchIndex(&command, outfile);
		}else{
			for(len = 0;len < ventries; ++len){
				AssembleCommand(&command, host[len], "GUTINDEX.ALL", "-");
				if(FetchIndex(&command, outfile))
					break;
			}
		}
	}else if(fetch_file != NULL)
		for(len = 0;len < ventries; ++len)
			if(FetchFile(fetch_file, outfile, host[len]))
				break;

	CleanUp(4, &fetch_file, &outfile, &force_host, &command);
	if(len < ventries)
		exit(TRUE);
	else
		exit(FALSE);
}

int
FetchFile(fetch_file, outfile, host)
	char *fetch_file;
	char *outfile;
	char *host;
{
	FILE *fp = NULL;
#ifdef HAVE_MKSTEMP
	int fpn = NULL;
#endif 
	int quit = FALSE;
	ARRAY buffer = { 0, NULL };
	char *basename = NULL;
	char *cmd = NULL;
	char *directory = NULL;
	char *filename = NULL;
	char *path = NULL;

	// get the basename of the file we are searching for
	if(strlen(fetch_file) > 8){
		if(verbose)
			fprintf(stderr, 
			"gutenfetch.c : FetchFile() : File %s must be fewer than 8 characters\n", fetch_file);
		return FALSE;
	}
	if((basename = malloc(sizeof(char) * (strlen(fetch_file) + 1))) == NULL){
		if(verbose)
			fprintf(stderr, "gutenfetch.c : FetchFile() : Memory allocation error\n");
		return FALSE;
	}
	strcpy(basename, fetch_file);

	// fetch the file ls-R  a listing of all files
	AssembleCommand(&cmd, host, "ls-R", "-");
	if(!OpenPipe(&fp, &cmd, "r")){
		CleanUp(2, &cmd, &basename);
		return FALSE;
	}

	// sort out interesting hits
	while((filename = getline(filename, fp)) != NULL){
		if(strncmp(filename, "./etext", 7) == 0){
			CleanUp(1, &directory);
			if(asprintf(&directory, "%s", &filename[2]) == -1){
				if(verbose)
					fprintf(stderr,
					"gutenfetch.c : FetchFile() : Memory Allocation error\n");
				CleanUp(2, &cmd, &basename);
				return FALSE;
			}
			directory[7] = '\0';
		}else if(strncmp(filename, basename, strlen(basename)) == 0)
			if(strncmp(&filename[strlen(filename)-3], "zip", 3) == 0)
				quit = TRUE;
		if(quit)
			break;
	}

	// close wget
	if(!ClosePipe(&fp)){
		CleanUp(4, &cmd, &basename, &filename, &directory);
		return FALSE;
	}

	// now we have a valid directory / filename to the file to read
	if(asprintf(&path, "%s/%s", directory, filename) == -1){
		if(verbose)
			fprintf(stderr,
			"gutenfetch.c : FetchFile() : Memory allocation error\n");
		CleanUp(4, &cmd, &basename, &filename, &directory);
		return FALSE;
	}
	AssembleCommand(&cmd, host, path, "-");
	// open pipe and read stdout of wget into buffer then close wget
	if(!OpenPipe(&fp, &cmd, "r")){
		CleanUp(5, &cmd, &basename, &filename, &directory, &path);
		return FALSE;
	}
	if(!ReadFile(fp, &buffer)){
		CleanUp(5, &cmd, &basename, &filename, &directory, &path);
		CleanUpArray(1, &buffer);
		return FALSE;
	}
	if(!ClosePipe(&fp)){
		CleanUp(5, &cmd, &basename, &filename, &directory, &path);
		CleanUpArray(1, &buffer);
		return FALSE;
	}
	CleanUp(2, &cmd, &filename);

	// write buffer to a temporary file and unzip it
	// writing the result to stdout or a file
	if(asprintf(&filename, "/tmp/gfetch.XXXXX") == -1){
		if(verbose)
			fprintf(stderr,
			"gutenfetch.c : FetchFile() : memory allocation error\n");
		CleanUp(3, &basename, &directory, &path);
		CleanUpArray(1, &buffer);
		return FALSE;
	}

#ifndef HAVE_MKSTEMP
	filename = mktemp(filename);
	if((fp = fopen(filename, "w")) == NULL){
#else /* HAVE_MKSTEMP is defined */
	if((fpn = mkstemp(filename)) == -1){
#endif /* HAVE_MKSTEMP */
		if(verbose)
			fprintf(stderr,
			"gutenfetch.c : FetchFile() : unable to open %s\n", filename);
		remove(filename);
		CleanUp(4, &basename, &directory, &path, &filename);
		CleanUpArray(1, &buffer);
		return FALSE;
	}

#ifndef HAVE_MKSTEMP	
	if(!WriteFile(fp, &buffer)){
		fclose(fp);
#else /* HAVE_MKSTEMP is defined */
	if(!WriteFile_fd(fpn, &buffer)){
		close(fpn);
#endif /* HAVE_MKSTEMP */
		remove(filename);
		CleanUp(4, &basename, &directory, &path, &filename);
		CleanUpArray(1, &buffer);
		return FALSE;
	}

#ifndef HAVE_MKSTEMP
	fclose(fp);
#else /* HAVE_MKSTEMP is defined */
	close(fpn);
#endif /* HAVE_MKSTEMP */

	if(outfile[0] == '-'){
		if(asprintf(&cmd, "unzip -p %s", filename) == -1){
			if(verbose)
				fprintf(stderr,
				"gutenfetch.c : FetchFile() : Memory allocation error\n");
			remove(filename);
			CleanUp(5, &cmd, &basename, &filename, &directory, &path);
			CleanUpArray(1, &buffer);
			return FALSE;
		}
	}else{
		if(asprintf(&cmd, "unzip -p %s > %s", filename, outfile) == -1){
			if(verbose)
				fprintf(stderr,
				"gutenfetch.c : FetchFile() : Memory allocation error\n");
			remove(filename);
			CleanUp(5, &cmd, &basename, &filename, &directory, &path);
			CleanUpArray(1, &buffer);
			return FALSE;
		}
	}

	// unzip the file
	system(cmd);
	// remove temporary file
	remove(filename);

	// free our memory
	CleanUp(5, &cmd, &basename, &filename, &directory, &path);
	CleanUpArray(1, &buffer);
	return TRUE;
}

char *
getline(line, fp)
	char *line;
	FILE *fp;
{
	int i = 0;
// I don't think this is a very good algorithm, 
// perhaps a more eleqeant solution could be
// used . . . 

	CleanUp(1, &line);
	if((line = malloc(sizeof(char) * 1)) == NULL){
		if(verbose)
			fprintf(stderr,
			"gutenfetch.c : getline() : Memory allocation error\n");
		return NULL;
	}
	while(TRUE){
		line[i] = fgetc(fp);
		if(line[i] == '\n'){
			line[i] = '\0';	
			break;
		}
		if(line[i] == EOF)
			return NULL;
		++i;
		if((line = realloc(line, sizeof(char) * (1 + i))) == NULL){
			if(verbose)
				fprintf(stderr,
				"gutenfetch.c : getline() : Memory allocation error\n");
			return NULL;
		}
	}
	return line;
}

int
FetchIndex(cmd, outfile)
	char **cmd;
	char *outfile;
{
	FILE *fp;
	ARRAY buffer = { 0, NULL };

	// open a pipe to wget
	if(!OpenPipe(&fp, cmd, "r"))
		return FALSE;

	// read the index file off gutenberg server into the buffer
	ReadFile(fp, &buffer);

	// close wget
	if(!ClosePipe(&fp))
		return FALSE;

	// pass the index file through grep and delete the non-book entries
	// this method isn't perfect but should suffice for the time being
	CleanUp(1, cmd);
	if(asprintf(cmd, "grep -a '^[A-Z][a-z][a-z] [0-9][0-9][0-9][0-9] '") == -1){
		if(verbose)
			fprintf(stderr,
			"gutenfetch.c : FetchFile() : Memory allocation error\n");
		return FALSE;
	}	
	if(outfile[0] != '-'){
		strcat(*cmd, " > ");
		strcat(*cmd, outfile);
	}

	// open pipe to grep
	if(!OpenPipe(&fp, cmd, "w"))
		return FALSE;
	
	// write index to grep
	WriteFile(fp, &buffer);

	// close grep, it will take care of outputing results to stdout
	// or a file if neccessary
	if(!ClosePipe(&fp))
		return FALSE;

	return TRUE;
}

int
ReadFile(fp, array)
	FILE *fp;
	ARRAY *array;
{
	int retval;

	// get initial space to work with
	if(!AllocateArray(array, sizeof(char) * BLOCK_SIZE))
		return FALSE;

	// read the stream in blocks
	while((retval = fread(&array->data[array->length-BLOCK_SIZE], sizeof(char), BLOCK_SIZE, fp)) == BLOCK_SIZE)
		if(!ReallocateArray(array, array->length + BLOCK_SIZE))
			return FALSE;

	// shave off unused memory
	if(!ReallocateArray(array, (array->length - (BLOCK_SIZE - retval))))
		return FALSE;

	return TRUE;
}

int
WriteFile(fp, array)
	FILE *fp;
	ARRAY *array;
{
	if((fp == NULL)||(array->data == NULL)||(array->length == 0)){
		if(verbose)
			fprintf(stderr,
			"gutenfetch.c : WriteFile() : Malformed ARRAY data structure\n");
		return FALSE;
	} 
	if(fwrite(array->data, sizeof(char), array->length, fp) < array->length){
		if(verbose)
			fprintf(stderr,
			"gutenfetch.c : WriteFile() : Error during write\n");
		return FALSE;
	}
	return TRUE;
}

#ifdef HAVE_MKSTEMP
int
WriteFile_fd(fd, array)
	int fd;
	ARRAY *array;
{
	if((fd == -1)||(array->data == NULL)||(array->length == 0)){
		if(verbose)
			fprintf(stderr,
			"gutenfetch.c : WriteFile() : Malformed ARRAY data structure\n");
		return FALSE;
	} 
	if(write(fd, array->data, (array->length * sizeof(char))) == -1){
		if(verbose)
			perror("gutenfetch.c : WriteFile_fd() : error writing data\n");
		return FALSE;
	}
	return TRUE;
}
#endif /* HAVE_MKSTEMP */

int 
AllocateArray(array, bytes)
	ARRAY *array;
	int bytes;
{
	CleanUpArray(1, array);
	if((array->data = malloc(bytes)) == NULL){
		if(verbose)
			fprintf(stderr,
			"gutenfetch.c : AllocateArray() : %d bytes failed to be allocated\n", bytes);
		return FALSE;
	}
	array->length = bytes;
	return TRUE;
}

int
ReallocateArray(array, bytes)
	ARRAY *array;
	int bytes;
{
	if((array->data = realloc(array->data, bytes)) == NULL){
		if(verbose)
			fprintf(stderr,
			"gutenfetch.c : ReallocateArray() : %d bytes failed to be allocated\n", bytes);
		array->length = 0;
		return FALSE;
	}
	array->length = bytes;
	return TRUE;
}


int
OpenPipe(fp, cmd, type)
	FILE **fp;
	char **cmd;
	char *type;
{
	if((*fp = popen(*cmd, type)) == NULL){
		if(verbose)
			fprintf(stderr,
			"gutenfetch.c : OpenPipe() : Unable to spawn external helper\n%s\n", *cmd);
		CleanUp(1, cmd);
		return FALSE;
	}
	return TRUE;
}

int ClosePipe(fp)
	FILE **fp;
{
	// finish reading stdout will cause an error otherwise
	while(fgetc(*fp) != EOF);
	if(pclose(*fp) != 0){
		if(verbose)
			fprintf(stderr,
			"gutenfetch.c : ClosePipe() : Error in external helper program\n");
		return FALSE;
	}
	return TRUE;
}

void
AssembleCommand(cmd, srv, inf, outf) 
	char **cmd;
	char *srv;
	char *inf;
	char *outf;
{
	CleanUp(1, cmd);
	asprintf(cmd, "wget -q -O %s %s%s", outf, srv, inf);
}
	
void
CleanUp(int i, ...)
{
	void **ptr;
	va_list ap;

	va_start(ap, i);
	for(;i>0;--i){
		ptr = va_arg(ap, void**);
		if((void*)*ptr != NULL){
			free(*ptr);	
			*ptr = NULL;
		}
	}
	va_end(ap); 
}

void 
CleanUpArray(int i, ...)
{
	void *ptr;
	va_list ap;

	va_start(ap, i);
	for(;i>0;--i){
		ptr = va_arg(ap, void*);
		((ARRAY*)ptr)->length = 0;
		if(((ARRAY*)ptr)->data != NULL){
			free(((ARRAY*)ptr)->data);
			((ARRAY*)ptr)->data = NULL;
		}
	}
	va_end(ap);
}

void 
PrintUsage(void)
{
	fprintf(stdout, "%s %s by %s\n", PACKAGE,  VERSION, "Russell Francis");
	fprintf(stdout, "usage: %s [-h] [-v] [-q] [-l] [-o file]", PACKAGE); 
	fprintf(stdout, "[-s server] -f ifile0\n");
	fprintf(stdout, "-h -\tThis help screen\n");
	fprintf(stdout, "-v -\tverbose, turn on all warnings and error messages\n");
	fprintf(stdout, "-q -\tsilent, turn off all warnings and error messages\n");
	fprintf(stdout, "-o -\tdirect output to a file, '-' = stdout, default is stdout\n");
	fprintf(stdout, "-s -\tforce the use of a particular gutenberg server\n");
	fprintf(stdout, "-l -\tlist availiable books\n");
	fprintf(stdout, "-f -\tfetch a particular book off of the server\n");
	fprintf(stdout, "    \talice - Alice in Wonderland output in ASCII text\n");
	fprintf(stdout, "    \tmanif - Communist Manifesto\n");
}

int
LoadConfigFile(void)
{
	int retval = FALSE;
	FILE *fp = NULL;
	char *home = NULL;
	char *filename = NULL;

	// test if ~/.gutenfetch exists
	home = getenv("HOME");
	if(asprintf(&filename, "%s/.gutenfetchrc", home) == -1){
		if(verbose)
			fprintf(stderr,
			"gutenfetch.c : LoadConfigFile() : Memory allocation error\n");
		return FALSE;
	}
	// try ~/.gutenfetchrc
	if((fp = fopen(filename, "r")) != NULL){
		if(!ParseConfigFile(fp)){
			if(verbose){
				fprintf(stderr, "gutenfetch.c : LoadConfigFile() :");
				fprintf(stderr, "Unable to read %s\n", filename);
			}
			retval = FALSE;
		}else
			retval = TRUE;
	// try /etc/gutenfetchrc
	}else if((fp = fopen("/etc/gutenfetchrc", "r")) != NULL){
		if(!ParseConfigFile(fp)){
			if(verbose){
				fprintf(stderr, "gutenfetch.c : LoadConfigFile() :");
				fprintf(stderr, "Unable to read /etc/gutenfetchrc\n");
			}
			retval = FALSE;
		}else
			retval = TRUE;
	// Build a new ~/.gutenfetchrc file
	}else{
		if(verbose){
			fprintf(stderr, "gutenfetch.c : LoadConfigFile() :");
			fprintf(stderr, "Trying to build new config file\n");
		}
		if((fp = fopen(filename, "w")) != NULL){
			GenerateNewConfigFile(fp);
			if(!ParseConfigFile(fp)){
				if(verbose){
					fprintf(stderr, "gutenfetch.c : LoadConfigFile() :");
					fprintf(stderr, "Unable to read %s\n", filename);
				}
				retval = FALSE;	
			}else	
				retval = TRUE;
		}else if(verbose){
			fprintf(stderr, "gutenfetch.c : LoadConfigFile() :");
			fprintf(stderr, "Unable to open config file\n");
		}
	}
	fclose(fp);
	CleanUp(1, &filename);
	return retval;
}

int
GenerateNewConfigFile(fp)
	FILE *fp;
{
	int i, j;
	int key = 'a';
	int cont;
	int cont_count;

	while(TRUE){
		for(key = 'a'; ((key < '0')||(key > '5')); key = fgetc(stdin)){
			fprintf(stdout, "\n[0] - North America\n");
			fprintf(stdout, "[1] - South America\n");
			fprintf(stdout, "[2] - Europe\n");
			fprintf(stdout, "[3] - Africa\n");
			fprintf(stdout, "[4] - Austrailia\n");
			fprintf(stdout, "[5] - Asia\n");		 
			fprintf(stdout, "Please choose a continent: ");
			fpurge(stdin);
		}

		switch(key){
			case '0':
				cont = CONTINENT_NORTH_AMERICA;
				break;
			case '1':
				cont = CONTINENT_SOUTH_AMERICA;
				break;
			case '2':
				cont = CONTINENT_EUROPE;
				break;
			case '3':
				cont = CONTINENT_AFRICA;
				break;
			case '4':
				cont = CONTINENT_AUSTRAILIA;
				break;
			case '5':
				cont = CONTINENT_ASIA;
				break;
			default:
				if(verbose)
					fprintf(stderr, "gutenfetch.c : \
					GenerateNewConfigFile() : \
					Something has gone terribly wrong!!!\n");
				return FALSE;
		}
		for(i = 0, cont_count = 0;i < (sizeof(servers) / sizeof(SERVER));++i)
			if(servers[i].continent == cont)
				cont_count++;

		for(key = -1;((key < 0)||(key > cont_count));){
			for(i = 0, j = 0;i < (sizeof(servers) / sizeof(SERVER));++i)
				if(servers[i].continent == cont)
					fprintf(stdout, "[%d] - %s - %s\n", 
					j++, servers[i].name, servers[i].area); 
			fprintf(stdout, "Select a server:  ");
			fpurge(stdin);
			fscanf(stdin, "%d", &key);	
		}

		for(i = 0, key++, cont_count = 0;i < (sizeof(servers) / sizeof(SERVER));++i){
			if(servers[i].continent == cont)
				cont_count++;
			if(cont_count == key)
				break;
		}
		fprintf(fp, "server %s\n", servers[i].host);
		fprintf(stdout, "Would you like to add another default server [Y/n]?  ");
		fpurge(stdin);
		key = fgetc(stdin);
		if((key == 'n')||(key == 'N'))
			break; 
	}
	return TRUE;
}

int
ParseConfigFile(fp)
	FILE *fp;
{
	char *tmp = NULL;
	char *action = NULL;
	char **ap = NULL;
	char *argv[2];

	while(TRUE){
		tmp = action = getline(tmp, fp);
		if(action == NULL){
			CleanUp(2, &action, &tmp);
			return TRUE;
		}
		for(ap = argv;(*ap = strsep(&action, " ")) != NULL;)
			if(**ap != '\0')
				if(++ap >= &argv[2])
					break;

		if(strncmp("server", argv[0], strlen("server")) == 0){
			ventries++;
			if((host = realloc(host, (sizeof(char *) * ventries))) == NULL){
				if(verbose)
					fprintf(stderr, 
					"gutenfetch.c : ParseConfigFile() : Memory allocation error\n");
				CleanUp(2, &action, &tmp);
				return FALSE;
			}
			if((host[(ventries - 1)] = malloc(sizeof(char) * (strlen(argv[1])+1))) == NULL){
				if(verbose)
					fprintf(stderr, 
					"gutenfetch.c : ParseConfigFile() : Memory allocation error\n");
				CleanUp(2, &action, &tmp);
				return FALSE;
			}
			strncpy(host[(ventries - 1)], argv[1], (strlen(argv[1]) + 1));
		}
		if(strncmp("verbose", argv[0], strlen("verbose")) == 0)
			verbose = atoi(argv[1]);
	}
	CleanUp(2, &action, &tmp);
	return TRUE;
}
